@@ -3,7 +3,6 @@ import base64
 import copy
 import functools
 import json
-import logging
 import os
 from datetime import timedelta
 
@@ -13,12 +12,13 @@ except ImportError:
     np = None
 
 import websockets
-from sqlitedict import SqliteDict
 
 from . import buttons as btn
 from . import endpoints as ep
 from .exceptions import PyLGTVPairException, PyLGTVCmdException, PyLGTVCmdError, PyLGTVServiceNotFoundError
 from .handshake import REGISTRATION_MESSAGE
+from .storage_proto import StorageProto
+from .storage_sqlitedict import StorageSqliteDict
 
 if np:
     from . import cal_commands as cal
@@ -31,15 +31,12 @@ if np:
         unity_lut_3d,
     )
 
-logger = logging.getLogger(__name__)
-
-KEY_FILE_NAME = ".aiopylgtv.sqlite"
-USER_HOME = "HOME"
-
 SOUND_OUTPUTS_TO_DELAY_CONSECUTIVE_VOLUME_STEPS = {"external_arc"}
 
 
 class WebOsClient:
+    STATIC_STATES = {"system_info", "software_info"}
+
     def __init__(
         self,
         ip,
@@ -49,14 +46,13 @@ class WebOsClient:
         ping_timeout=20,
         client_key=None,
         volume_step_delay_ms=None,
-        getSystemInfo=True,
-        skipStateInfo=False,
-        disable_key_file=False,
+        states=["system_info", "software_info", "power", "current_app", "muted",
+                "volume", "apps", "inputs", "sound_output", "picture_settings"],
+        storage: StorageProto=None,
     ):
         """Initialize the client."""
         self.ip = ip
         self.port = 3000
-        self._disable_key_file = disable_key_file
         self.key_file_path = key_file_path
         self.client_key = client_key
         self.web_socket = None
@@ -64,8 +60,7 @@ class WebOsClient:
         self.timeout_connect = timeout_connect
         self.ping_interval = ping_interval
         self.ping_timeout = ping_timeout
-        self.getSystemInfo = getSystemInfo
-        self.skipStateInfo = skipStateInfo
+        self.storage = storage
         self.connect_task = None
         self.connect_result = None
         self.connection = None
@@ -94,6 +89,7 @@ class WebOsClient:
             if volume_step_delay_ms is not None
             else None
         )
+        self.states = (set(states) if isinstance(states, list) else set())
 
     @classmethod
     async def create(cls, *args, **kwargs):
@@ -102,62 +98,17 @@ class WebOsClient:
         return client
 
     async def async_init(self):
-        """Load client key from config file if in use."""
-        if self._disable_key_file:
-            return
-
+        """Load client key from storage if it's required."""
         if self.client_key is None:
-            self.client_key = await asyncio.get_running_loop().run_in_executor(
-                None, self.read_client_key
-            )
+            if self.storage is None:
+                self.storage = await StorageSqliteDict.create(self.key_file_path)
+            elif not isinstance(self.storage, StorageProto):
+                raise PyLGTVCmdException("Storage is not a StorageProto class.")
 
-    @staticmethod
-    def _get_key_file_path():
-        """Return the key file path."""
-        if os.getenv(USER_HOME) is not None and os.access(
-            os.getenv(USER_HOME), os.W_OK
-        ):
-            return os.path.join(os.getenv(USER_HOME), KEY_FILE_NAME)
+            self.client_key = await self.storage.get_key(self.ip)
 
-        return os.path.join(os.getcwd(), KEY_FILE_NAME)
-
-    def list_client_keys(key_file_path):
-        """Display all saved client keys per ip."""
-
-        logger.debug("load keyfile from %s", key_file_path)
-
-        with SqliteDict(key_file_path) as conf:
-            for ip, key in conf.iteritems():
-                print(ip, key)
-
-    def read_client_key(self):
-        """Try to load the client key for the current ip."""
-
-        if self.key_file_path:
-            key_file_path = self.key_file_path
-        else:
-            key_file_path = self._get_key_file_path()
-
-        logger.debug("load keyfile from %s", key_file_path)
-
-        with SqliteDict(key_file_path) as conf:
-            return conf.get(self.ip)
-
-    def write_client_key(self):
-        """Save the current client key."""
-        if self.client_key is None:
-            return
-
-        if self.key_file_path:
-            key_file_path = self.key_file_path
-        else:
-            key_file_path = self._get_key_file_path()
-
-        logger.debug("save keyfile to %s", key_file_path)
-
-        with SqliteDict(key_file_path) as conf:
-            conf[self.ip] = self.client_key
-            conf.commit()
+    async def get_storage(self):
+        return self.storage
 
     async def connect(self):
         if not self.is_connected():
@@ -212,10 +163,7 @@ class WebOsClient:
                 response = json.loads(raw_response)
                 if response["type"] == "registered":
                     self.client_key = response["payload"]["client-key"]
-                    if not self._disable_key_file:
-                        await asyncio.get_running_loop().run_in_executor(
-                            None, self.write_client_key
-                        )
+                    await self.storage.set_key(self.ip, self.client_key)
 
             if not self.client_key:
                 raise PyLGTVPairException("Unable to pair")
@@ -237,35 +185,38 @@ class WebOsClient:
             self.connection = ws
 
             self.doStateUpdate = False
-            if self.getSystemInfo:
-                # set static state
-                [self._system_info] = await asyncio.gather(self.get_system_info())
+            if self.states:
+                # set static states, possible values: ["system_info", "software_info"]
+                staticStates = self.states.intersection(self.STATIC_STATES)
+                if staticStates:
+                    # e.g.: [self._system_info] = await asyncio.gather(self.get_system_info())
+                    for stateElem in staticStates:
+                        stateResult = await asyncio.gather(getattr(self, f'get_{stateElem}')())
+                        setattr(self, f'_{stateElem}', (stateResult or [None])[0])
+                        self.states.remove(stateElem)
 
-            if not self.skipStateInfo:
-                # set static state and subscribe to state updates
-                # avoid partial updates during initial subscription
-                [self._software_info] = await asyncio.gather(self.get_software_info())
+                # subscribe to state updates, avoid partial updates during initial subscription
+                # possible values: ["power", "current_app", "muted", "volume", "apps", "inputs",
+                #                   "sound_output", "picture_settings"]
+                subscribe_coros = set()
+                if self.states:
+                    # e.g.: subscribe_coros.add(self.subscribe_power(self.set_power_state))
+                    for stateElem in self.states:
+                        subscriber = f'subscribe_{stateElem}'
+                        setter = f'set_{stateElem}_state'
+                        if callable(getattr(self, subscriber, None)) and callable(getattr(self, setter, None)):
+                            subscribe_coros.add(getattr(self, subscriber)(getattr(self, setter)))
 
-                subscribe_coros = {
-                    self.subscribe_power_state(self.set_power_state),
-                    self.subscribe_current_app(self.set_current_app_state),
-                    self.subscribe_muted(self.set_muted_state),
-                    self.subscribe_volume(self.set_volume_state),
-                    self.subscribe_apps(self.set_apps_state),
-                    self.subscribe_inputs(self.set_inputs_state),
-                    self.subscribe_sound_output(self.set_sound_output_state),
-                    self.subscribe_picture_settings(self.set_picture_settings_state),
-                }
-
-                subscribe_tasks = set()
-                for coro in subscribe_coros:
-                    subscribe_tasks.add(asyncio.create_task(coro))
-                await asyncio.wait(subscribe_tasks)
-                for task in subscribe_tasks:
-                    try:
-                        task.result()
-                    except PyLGTVServiceNotFoundError:
-                        pass
+                if subscribe_coros:
+                    subscribe_tasks = set()
+                    for coro in subscribe_coros:
+                        subscribe_tasks.add(asyncio.create_task(coro))
+                    await asyncio.wait(subscribe_tasks)
+                    for task in subscribe_tasks:
+                        try:
+                            task.result()
+                        except (PyLGTVCmdError, PyLGTVServiceNotFoundError):
+                            pass
 
             # set placeholder power state if not available
             if not self._power_state:
@@ -302,6 +253,7 @@ class WebOsClient:
 
             self.doStateUpdate = False
 
+            self.storage = None
             self._power_state = {}
             self._current_appId = None
             self._muted = None
@@ -317,7 +269,7 @@ class WebOsClient:
             self._picture_settings = None
 
             for callback in self.state_update_callbacks:
-                closeout.add(callback())
+                closeout.add(callback(self))
 
             if closeout:
                 closeout_task = asyncio.create_task(asyncio.wait(closeout))
@@ -473,7 +425,7 @@ class WebOsClient:
     async def register_state_update_callback(self, callback):
         self.state_update_callbacks.append(callback)
         if self.doStateUpdate:
-            await callback()
+            await callback(self)
 
     def unregister_state_update_callback(self, callback):
         if callback in self.state_update_callbacks:
@@ -485,13 +437,13 @@ class WebOsClient:
     async def do_state_update_callbacks(self):
         callbacks = set()
         for callback in self.state_update_callbacks:
-            callbacks.add(callback())
+            callbacks.add(callback(self))
 
         if callbacks:
             await asyncio.gather(*callbacks)
 
     async def set_power_state(self, payload):
-        self._power_state = payload
+        self._power_state = {"state": payload.get("state", "Unknown")}
 
         if self.state_update_callbacks and self.doStateUpdate:
             await self.do_state_update_callbacks()
@@ -585,7 +537,10 @@ class WebOsClient:
             await self.do_state_update_callbacks()
 
     async def set_picture_settings_state(self, picture_settings):
-        self._picture_settings = picture_settings
+        if isinstance(self._picture_settings, dict) and isinstance(picture_settings, dict):
+            self._picture_settings.update(picture_settings)
+        else:
+            self._picture_settings = picture_settings
 
         if self.state_update_callbacks and self.doStateUpdate:
             await self.do_state_update_callbacks()
@@ -749,7 +704,7 @@ class WebOsClient:
         """Get current power state."""
         return await self.request(ep.GET_POWER_STATE)
 
-    async def subscribe_power_state(self, callback):
+    async def subscribe_power(self, callback):
         """Subscribe to current power state."""
         return await self.subscribe(callback, ep.GET_POWER_STATE)
 
@@ -815,7 +770,8 @@ class WebOsClient:
         """Power off TV."""
 
         # protect against turning tv back on if it is off
-        self._power_state = await self.get_power_state()
+        power_state = await self.get_power_state()
+        self._power_state = {"state": power_state.get("state", "Unknown")}
         if not self.is_on:
             return
 
